//
//    Copyright (C) 2018 Sascha Ittner <sascha.ittner@modusoft.de>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
//

#include "lcec.h"
#include "lcec_el1819.h"

typedef struct {
  hal_u32_t *fsoe_master_cmd;
  hal_u32_t *fsoe_master_crc;
  hal_u32_t *fsoe_master_connid;

  hal_u32_t *fsoe_slave_cmd;
  hal_u32_t *fsoe_slave_crc;
  hal_u32_t *fsoe_slave_connid;

  hal_bit_t *fsoe_in_0;
  hal_bit_t *fsoe_in_1;
  hal_bit_t *fsoe_in_2;
  hal_bit_t *fsoe_in_3;
  hal_bit_t *fsoe_in_4;
  hal_bit_t *fsoe_in_5;
  hal_bit_t *fsoe_in_6;
  hal_bit_t *fsoe_in_7;
  hal_bit_t *fsoe_in_8;
  hal_bit_t *fsoe_in_9;
  hal_bit_t *fsoe_in_10;
  hal_bit_t *fsoe_in_11;
  hal_bit_t *fsoe_in_12;
  hal_bit_t *fsoe_in_13;
  hal_bit_t *fsoe_in_14;
  hal_bit_t *fsoe_in_15;

  unsigned int fsoe_master_cmd_os;
  unsigned int fsoe_master_crc_os;
  unsigned int fsoe_master_connid_os;

  unsigned int fsoe_slave_cmd_os;
  unsigned int fsoe_slave_crc_os;
  unsigned int fsoe_slave_connid_os;

  unsigned int fsoe_in_0_os;
  unsigned int fsoe_in_0_bp;
  unsigned int fsoe_in_1_os;
  unsigned int fsoe_in_1_bp;
  unsigned int fsoe_in_2_os;
  unsigned int fsoe_in_2_bp;
  unsigned int fsoe_in_3_os;
  unsigned int fsoe_in_3_bp;
  unsigned int fsoe_in_4_os;
  unsigned int fsoe_in_4_bp;
  unsigned int fsoe_in_5_os;
  unsigned int fsoe_in_5_bp;
  unsigned int fsoe_in_6_os;
  unsigned int fsoe_in_6_bp;
  unsigned int fsoe_in_7_os;
  unsigned int fsoe_in_7_bp;
  unsigned int fsoe_in_8_os;
  unsigned int fsoe_in_8_bp;
  unsigned int fsoe_in_9_os;
  unsigned int fsoe_in_9_bp;
  unsigned int fsoe_in_10_os;
  unsigned int fsoe_in_10_bp;
  unsigned int fsoe_in_11_os;
  unsigned int fsoe_in_11_bp;
  unsigned int fsoe_in_12_os;
  unsigned int fsoe_in_12_bp;
  unsigned int fsoe_in_13_os;
  unsigned int fsoe_in_13_bp;
  unsigned int fsoe_in_14_os;
  unsigned int fsoe_in_14_bp;
  unsigned int fsoe_in_15_os;
  unsigned int fsoe_in_15_bp;

} lcec_el1819_data_t;

static const lcec_pindesc_t slave_pins[] = {
  { HAL_U32, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_master_cmd), "%s.%s.%s.fsoe-master-cmd" },
  { HAL_U32, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_master_crc), "%s.%s.%s.fsoe-master-crc" },
  { HAL_U32, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_master_connid), "%s.%s.%s.fsoe-master-connid" },
  { HAL_U32, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_slave_cmd), "%s.%s.%s.fsoe-slave-cmd" },
  { HAL_U32, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_slave_crc), "%s.%s.%s.fsoe-slave-crc" },
  { HAL_U32, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_slave_connid), "%s.%s.%s.fsoe-slave-connid" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_0), "%s.%s.%s.fsoe-in-0" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_1), "%s.%s.%s.fsoe-in-1" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_2), "%s.%s.%s.fsoe-in-2" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_3), "%s.%s.%s.fsoe-in-3" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_4), "%s.%s.%s.fsoe-in-4" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_5), "%s.%s.%s.fsoe-in-5" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_6), "%s.%s.%s.fsoe-in-6" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_7), "%s.%s.%s.fsoe-in-7" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_8), "%s.%s.%s.fsoe-in-8" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_9), "%s.%s.%s.fsoe-in-9" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_10), "%s.%s.%s.fsoe-in-10" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_11), "%s.%s.%s.fsoe-in-11" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_12), "%s.%s.%s.fsoe-in-12" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_13), "%s.%s.%s.fsoe-in-13" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_14), "%s.%s.%s.fsoe-in-14" },
  { HAL_BIT, HAL_OUT, offsetof(lcec_el1819_data_t, fsoe_in_15), "%s.%s.%s.fsoe-in-15" },
  { HAL_TYPE_UNSPECIFIED, HAL_DIR_UNSPECIFIED, -1, NULL }
};

void lcec_el1819_read(struct lcec_slave *slave, long period);

int lcec_el1819_init(int comp_id, struct lcec_slave *slave, ec_pdo_entry_reg_t *pdo_entry_regs) {
  lcec_master_t *master = slave->master;
  lcec_el1819_data_t *hal_data;
  int err;

  // initialize callbacks
  slave->proc_read = lcec_el1819_read;

  // alloc hal memory
  if ((hal_data = hal_malloc(sizeof(lcec_el1819_data_t))) == NULL) {
    rtapi_print_msg(RTAPI_MSG_ERR, LCEC_MSG_PFX "hal_malloc() for slave %s.%s failed\n", master->name, slave->name);
    return -EIO;
  }
  memset(hal_data, 0, sizeof(lcec_el1819_data_t));
  slave->hal_data = hal_data;

  // initialize POD entries
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x7000, 0x01, &hal_data->fsoe_master_cmd_os, NULL);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x7000, 0x02, &hal_data->fsoe_master_crc_os, NULL);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x7000, 0x03, &hal_data->fsoe_master_connid_os, NULL);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x7000, 0x04, &hal_data->fsoe_slave_cmd_os, NULL);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x01, &hal_data->fsoe_in_0_os, &hal_data->fsoe_in_0_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x02, &hal_data->fsoe_in_1_os, &hal_data->fsoe_in_1_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x03, &hal_data->fsoe_in_2_os, &hal_data->fsoe_in_2_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x04, &hal_data->fsoe_in_3_os, &hal_data->fsoe_in_3_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x05, &hal_data->fsoe_in_4_os, &hal_data->fsoe_in_4_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x06, &hal_data->fsoe_in_5_os, &hal_data->fsoe_in_5_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x07, &hal_data->fsoe_in_6_os, &hal_data->fsoe_in_6_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x08, &hal_data->fsoe_in_7_os, &hal_data->fsoe_in_7_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x09, &hal_data->fsoe_in_8_os, &hal_data->fsoe_in_8_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x10, &hal_data->fsoe_in_9_os, &hal_data->fsoe_in_9_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x11, &hal_data->fsoe_in_10_os, &hal_data->fsoe_in_10_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x12, &hal_data->fsoe_in_11_os, &hal_data->fsoe_in_11_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x13, &hal_data->fsoe_in_12_os, &hal_data->fsoe_in_12_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x14, &hal_data->fsoe_in_13_os, &hal_data->fsoe_in_13_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x15, &hal_data->fsoe_in_14_os, &hal_data->fsoe_in_14_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6001, 0x16, &hal_data->fsoe_in_15_os, &hal_data->fsoe_in_15_bp);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6000, 0x03, &hal_data->fsoe_slave_crc_os, NULL);
  LCEC_PDO_INIT(pdo_entry_regs, slave->index, slave->vid, slave->pid, 0x6000, 0x04, &hal_data->fsoe_slave_connid_os, NULL);

  // export pins
  if ((err = lcec_pin_newf_list(hal_data, slave_pins, LCEC_MODULE_NAME, master->name, slave->name)) != 0) {
    return err;
  }

  return 0;
}

void lcec_el1819_read(struct lcec_slave *slave, long period) {
  lcec_master_t *master = slave->master;
  lcec_el1819_data_t *hal_data = (lcec_el1819_data_t *) slave->hal_data;
  uint8_t *pd = master->process_data;

  copy_fsoe_data(slave, hal_data->fsoe_slave_cmd_os, hal_data->fsoe_master_cmd_os);

  *(hal_data->fsoe_slave_cmd) = EC_READ_U8(&pd[hal_data->fsoe_slave_cmd_os]);
  *(hal_data->fsoe_slave_crc) = EC_READ_U16(&pd[hal_data->fsoe_slave_crc_os]);
  *(hal_data->fsoe_slave_connid) = EC_READ_U16(&pd[hal_data->fsoe_slave_connid_os]);

  *(hal_data->fsoe_master_cmd) = EC_READ_U8(&pd[hal_data->fsoe_master_cmd_os]);
  *(hal_data->fsoe_master_crc) = EC_READ_U16(&pd[hal_data->fsoe_master_crc_os]);
  *(hal_data->fsoe_master_connid) = EC_READ_U16(&pd[hal_data->fsoe_master_connid_os]);

  *(hal_data->fsoe_in_0) = EC_READ_BIT(&pd[hal_data->fsoe_in_0_os], hal_data->fsoe_in_0_bp);
  *(hal_data->fsoe_in_1) = EC_READ_BIT(&pd[hal_data->fsoe_in_1_os], hal_data->fsoe_in_1_bp);
  *(hal_data->fsoe_in_2) = EC_READ_BIT(&pd[hal_data->fsoe_in_2_os], hal_data->fsoe_in_2_bp);
  *(hal_data->fsoe_in_3) = EC_READ_BIT(&pd[hal_data->fsoe_in_3_os], hal_data->fsoe_in_3_bp);
  *(hal_data->fsoe_in_4) = EC_READ_BIT(&pd[hal_data->fsoe_in_4_os], hal_data->fsoe_in_4_bp);
  *(hal_data->fsoe_in_5) = EC_READ_BIT(&pd[hal_data->fsoe_in_5_os], hal_data->fsoe_in_5_bp);
  *(hal_data->fsoe_in_6) = EC_READ_BIT(&pd[hal_data->fsoe_in_6_os], hal_data->fsoe_in_6_bp);
  *(hal_data->fsoe_in_7) = EC_READ_BIT(&pd[hal_data->fsoe_in_7_os], hal_data->fsoe_in_7_bp);
  *(hal_data->fsoe_in_8) = EC_READ_BIT(&pd[hal_data->fsoe_in_8_os], hal_data->fsoe_in_8_bp);
  *(hal_data->fsoe_in_9) = EC_READ_BIT(&pd[hal_data->fsoe_in_9_os], hal_data->fsoe_in_9_bp);
  *(hal_data->fsoe_in_10) = EC_READ_BIT(&pd[hal_data->fsoe_in_10_os], hal_data->fsoe_in_10_bp);
  *(hal_data->fsoe_in_11) = EC_READ_BIT(&pd[hal_data->fsoe_in_11_os], hal_data->fsoe_in_11_bp);
  *(hal_data->fsoe_in_12) = EC_READ_BIT(&pd[hal_data->fsoe_in_12_os], hal_data->fsoe_in_12_bp);
  *(hal_data->fsoe_in_13) = EC_READ_BIT(&pd[hal_data->fsoe_in_13_os], hal_data->fsoe_in_13_bp);
  *(hal_data->fsoe_in_14) = EC_READ_BIT(&pd[hal_data->fsoe_in_14_os], hal_data->fsoe_in_14_bp);
  *(hal_data->fsoe_in_15) = EC_READ_BIT(&pd[hal_data->fsoe_in_15_os], hal_data->fsoe_in_15_bp);
}

